# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from threading import Thread
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.transceiver import create_transceiver_from_hostname
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from spinnman.connections.udp_packet_connections import EIEIOConnection
logger = FormatAdapter(logging.getLogger(__name__))


class MachineAllocationController(
        AbstractMachineAllocationController, metaclass=AbstractBase):
    """
    How to manage the allocation of a machine so that it gets cleaned up
    neatly when the script dies.
    """
    __slots__ = [
        #: boolean flag for telling this thread when the system has ended
        "_exited",
        #: the address of the root board of the allocation
        "__hostname",
        "__connection_data"
    ]

    def __init__(self, thread_name, hostname=None, connection_data=None):
        """
        :param str thread_name:
        """
        thread = Thread(name=thread_name, target=self.__manage_allocation)
        thread.daemon = True
        self._exited = False
        self.__hostname = hostname
        self.__connection_data = connection_data
        thread.start()

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        self._exited = True

    @abstractmethod
    def _wait(self):
        """
        Wait for some bounded amount of time for a change in the status
        of the machine allocation.

        :return: Whether the machine is still (believed to be) allocated.
        :rtype: bool
        """

    def _teardown(self):
        """
        Perform any extra tear-down that the thread requires. Does not
        need to be overridden if no action is desired.
        """

    def __manage_allocation(self):
        machine_still_allocated = True
        while machine_still_allocated and not self._exited:
            machine_still_allocated = self._wait()
        self._teardown()
        if not self._exited:
            logger.error(
                "The allocated machine has been released before the end of"
                " the script; this script will now exit")
            sys.exit(1)

    @overrides(AbstractMachineAllocationController.create_transceiver)
    def create_transceiver(self):
        if not self.__hostname:
            return None
        txrx = create_transceiver_from_hostname(
            hostname=self.__hostname,
            bmp_connection_data=None,
            version=5, auto_detect_bmp=False)
        txrx.ensure_board_is_ready()
        txrx.discover_scamp_connections()
        return txrx

    @overrides(AbstractMachineAllocationController.open_sdp_connection)
    def open_sdp_connection(self, chip_x, chip_y, udp_port=SCP_SCAMP_PORT):
        if not self.__connection_data:
            return None
        host = self.__connection_data[chip_x, chip_y]
        if not host:
            return None
        return SCAMPConnection(
            chip_x=chip_x, chip_y=chip_y,
            remote_host=host, remote_port=udp_port)

    @overrides(AbstractMachineAllocationController.open_eieio_connection)
    def open_eieio_connection(self, chip_x, chip_y):
        if not self.__connection_data:
            return None
        host = self.__connection_data[chip_x, chip_y]
        if not host:
            return None
        return EIEIOConnection(remote_host=host, remote_port=SCP_SCAMP_PORT)

    @overrides(AbstractMachineAllocationController.open_eieio_listener)
    def open_eieio_listener(self):
        return EIEIOConnection()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import sys
import time
from datetime import timedelta
from spinn_utilities.config_holder import (get_config_bool)
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView
from .global_provenance import GlobalProvenance

logger = FormatAdapter(logging.getLogger(__name__))

if sys.version_info >= (3, 7):
    # acquire the most accurate measurement available (perf_counter_ns)
    _now = time.perf_counter_ns  # pylint: disable=no-member
    # conversion factor
    _NANO_TO_MICRO = 1000.0

    def _convert_to_timedelta(time_diff):
        """
        Have to convert to a timedelta for rest of code to read.

        As perf_counter_ns is nano seconds, and time delta lowest is micro,
        need to convert.
        """
        return timedelta(microseconds=time_diff / _NANO_TO_MICRO)

else:
    # acquire the most accurate measurement available (perf_counter)
    _now = time.perf_counter  # pylint: disable=no-member

    def _convert_to_timedelta(time_diff):
        """
        Have to convert to a timedelta for rest of code to read.

        As perf_counter is fractional seconds, put into correct time delta.
        """
        return timedelta(seconds=time_diff)


class FecTimer(object):

    _simulator = None
    _provenance_path = None
    _print_timings = False
    _category_id = None
    _category = None
    _category_time = None
    _machine_on = False
    _previous = []
    __slots__ = [

        # The start time when the timer was set off
        "_start_time",

        # Name of algorithm what is being timed
        "_algorithm",

        # Type of work being done
        "_work"
        ]

    # Algorithm Names used elsewhere
    APPLICATION_RUNNER = "Application runner"

    @classmethod
    def setup(cls, simulator):
        # pylint: disable=global-statement, protected-access
        cls._simulator = simulator
        if get_config_bool("Reports", "write_algorithm_timings"):
            cls._provenance_path = os.path.join(
                FecDataView.get_run_dir_path(),
                "algorithm_timings.rpt")
        else:
            cls._provenance_path = None
        cls._print_timings = get_config_bool(
            "Reports", "display_algorithm_timings")

    def __init__(self, algorithm, work):
        self._start_time = None
        self._algorithm = algorithm
        self._work = work

    def __enter__(self):
        self._start_time = _now()
        return self

    def _report(self, message):
        if self._provenance_path is not None:
            with open(self._provenance_path, "a", encoding="utf-8") as p_file:
                p_file.write(f"{message}\n")
        if self._print_timings:
            logger.info(message)

    def skip(self, reason):
        message = f"{self._algorithm} skipped as {reason}"
        time_taken = self._stop_timer()
        with GlobalProvenance() as db:
            db.insert_timing(self._category_id, self._algorithm, self._work,
                             time_taken, reason)
        self._report(message)

    def skip_if_has_not_run(self):
        if self._simulator.has_ran:
            return False
        else:
            self.skip("simulator.has_run")
            return True

    def skip_if_virtual_board(self):
        if get_config_bool("Machine", "virtual_board"):
            self.skip("virtual_board")
            return True
        else:
            return False

    def skip_if_empty(self, value, name):
        if value:
            return False
        if value is None:
            self.skip(f"{name} is None")
        elif len(value) == 0:
            self.skip(f"{name} is empty")
        else:
            self.skip(f"{name} is False for an unknown reason")
        return True

    def skip_if_cfg_false(self, section, option):
        if get_config_bool(section, option):
            return False
        else:
            self.skip(f"cfg {section}:{option} is False")
            return True

    def skip_if_cfgs_false(self, section, option1, option2):
        if get_config_bool(section, option1):
            return False
        elif get_config_bool(section, option2):
            return False
        else:
            self.skip(f"cfg {section}:{option1} and {option2} are False")
            return True

    def error(self, reason):
        time_taken = self._stop_timer()
        message = f"{self._algorithm} failed after {timedelta} as {reason}"
        with GlobalProvenance() as db:
            db.insert_timing(self._category_id, self._algorithm,
                             self._work, time_taken, reason)
        self._report(message)

    def _stop_timer(self):
        """
        Describes how long has elapsed since the instance that the
        :py:meth:`start_timing` method was last called.

        :rtype: datetime.timedelta
        """
        time_now = _now()
        diff = time_now - self._start_time
        self._start_time = None
        return _convert_to_timedelta(diff)

    def __exit__(self, exc_type, exc_value, traceback):
        if self._start_time is None:
            return False
        time_taken = self._stop_timer()
        if exc_type is None:
            message = f"{self._algorithm} took {time_taken} "
            skip = None
        else:
            try:
                message = (f"{self._algorithm} exited with "
                           f"{exc_type.__name__} after {time_taken}")
                skip = exc_type.__name__
            except Exception as ex:  # pylint: disable=broad-except
                message = (f"{self._algorithm} exited with an exception"
                           f"after {time_taken}")
                skip = f"Exception {ex}"

        with GlobalProvenance() as db:
            db.insert_timing(self._category_id, self._algorithm, self._work,
                             time_taken, skip)
        self._report(message)
        return False

    @classmethod
    def __stop_category(cls):
        """
        Stops the current category and logs how long it took

        :return: Time the stop happened
        """
        time_now = _now()
        if cls._category_id:
            with GlobalProvenance() as db:
                diff = _convert_to_timedelta(time_now - cls._category_time)
                db.insert_category_timing(cls._category_id, diff)
        return time_now

    @classmethod
    def _change_category(cls, category):
        """
        This method should only be called via the View!

        :param TimerCategory category: Category to switch to
        """
        time_now = cls.__stop_category()
        with GlobalProvenance() as db:
            cls._category_id = db.insert_category(category, cls._machine_on)
        cls._category = category
        cls._category_time = time_now

    @classmethod
    def start_category(cls, category, machine_on=None):
        """
        This method should only be called via the View!

        :param TimerCategory category: category to switch to
        :param machine_on: What to change machine on too.
            Or `None` to leave as is
        :type machine_on: None or bool
        """
        cls._previous.append(cls._category)
        if cls._category != category:
            cls._change_category(category)
        if machine_on is not None:
            cls._machine_on = machine_on

    @classmethod
    def end_category(cls, category):
        """
        This method should only be
        called via the View!

        :param SimulatorStage category: Stage to end
        """
        if cls._category != category:
            raise ValueError(
                f"Current category is {cls._category} not {category}")
        previous = cls._previous.pop()
        if previous is None:
            raise NotImplementedError(
                "Use stop_category_timing to end the last category")
        if category != previous:
            cls._change_category(previous)

    @classmethod
    def stop_category_timing(cls):
        cls.__stop_category()
        cls._previous = []
        cls._category = None
        cls._category_id = None



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.utilities.iobuf_extractor import IOBufExtractor


def chip_io_buf_extractor():
    """
    Extract the logging output buffers from the machine, and separates
    lines based on their prefix.

    :return: error_entries, warn_entries
    :rtype: tuple(list(str),list(str))
    """
    extractor = IOBufExtractor()
    return extractor.extract_iobuf()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from spinn_utilities.log import FormatAdapter
from spinnman.messages.scp.enums import Signal
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION)

SAFETY_FINISH_TIME = 0.1

logger = FormatAdapter(logging.getLogger(__name__))


def application_runner(runtime, time_threshold, run_until_complete):
    """
    Ensures all cores are initialised correctly, ran, and completed
    successfully.

    :param int runtime:
    :param int time_threshold:
    :param bool run_until_complete:
    :raises ConfigurationException:
    """
    runner = _ApplicationRunner()
    # pylint: disable=protected-access
    runner._run(runtime, time_threshold, run_until_complete)


class _ApplicationRunner(object):
    """
    Ensures all cores are initialised correctly, ran, and completed
    successfully.
    """

    __slots__ = ["__txrx", "__app_id"]

    def __init__(self):
        self.__txrx = FecDataView.get_transceiver()
        self.__app_id = FecDataView.get_app_id()

    # Wraps up as a PACMAN algorithm
    def _run(
            self, runtime, time_threshold, run_until_complete=False):
        """
        :param int runtime:
        :param int time_threshold:
        :param bool run_until_complete:
        :return: Number of synchronisation changes
        :rtype: int
        :raises ConfigurationException:
        """
        # pylint: disable=too-many-arguments
        logger.info("*** Running simulation... *** ")

        # wait for all cores to be ready
        self._wait_for_start()

        buffer_manager = FecDataView.get_buffer_manager()
        notification_interface = FecDataView.get_notification_protocol()
        # set the buffer manager into a resume state, so that if it had ran
        # before it'll work again
        buffer_manager.resume()

        # every thing is in sync0 so load the initial buffers
        buffer_manager.load_initial_buffers()

        # clear away any router diagnostics that have been set due to all
        # loading applications
        for chip in FecDataView.get_machine().chips:
            self.__txrx.clear_router_diagnostic_counters(chip.x, chip.y)

        # wait till external app is ready for us to start if required
        notification_interface.wait_for_confirmation()

        # set off the executables that are in sync state
        # (sending to all is just as safe)
        self._send_sync_signal()

        # Send start notification to external applications
        notification_interface.send_start_resume_notification()

        if runtime is None and not run_until_complete:
            # Do NOT stop the buffer manager at end; app is using it still
            logger.info("Application is set to run forever; exiting")
        else:
            # Wait for the application to finish
            self._run_wait(
                run_until_complete, runtime, time_threshold)

            # Send stop notification to external applications
            notification_interface.send_stop_pause_notification()

    def _run_wait(self, run_until_complete, runtime, time_threshold):
        """
        :param bool run_until_complete:
        :param int runtime:
        :param float time_threshold:
        """
        if not run_until_complete:
            factor = (FecDataView.get_time_scale_factor() /
                      MICRO_TO_MILLISECOND_CONVERSION)
            scaled_runtime = runtime * factor
            time_to_wait = scaled_runtime + SAFETY_FINISH_TIME
            logger.info(
                "Application started; waiting {}s for it to stop",
                time_to_wait)
            time.sleep(time_to_wait)
            self._wait_for_end(timeout=time_threshold)
        else:
            logger.info("Application started; waiting until finished")
            self._wait_for_end()

    def _wait_for_start(self, timeout=None):
        """
        :param timeout:
        :type timeout: float or None
        """
        for ex_type, cores in FecDataView.get_executable_types().items():
            self.__txrx.wait_for_cores_to_be_in_state(
                cores, self.__app_id, ex_type.start_state, timeout=timeout)

    def _send_sync_signal(self):
        """
        Let apps that use the simulation interface or sync signals commence
        running their main processing loops. This is done with a very fast
        synchronisation barrier and a signal.
        """
        executable_types = FecDataView.get_executable_types()
        if (ExecutableType.USES_SIMULATION_INTERFACE in executable_types
                or ExecutableType.SYNC in executable_types):
            # locate all signals needed to set off executables
            sync_signal = self._determine_simulation_sync_signals()

            # fire all signals as required
            self.__txrx.send_signal(self.__app_id, sync_signal)

    def _wait_for_end(self, timeout=None):
        """
        :param timeout:
        :type timeout: float or None
        """
        for ex_type, cores in FecDataView.get_executable_types().items():
            self.__txrx.wait_for_cores_to_be_in_state(
                cores, self.__app_id, ex_type.end_state, timeout=timeout)

    def _determine_simulation_sync_signals(self):
        """
        Determines the start states, and creates core subsets of the
        states for further checks.

        :return: the sync signal
        :rtype: ~.Signal
        :raises ConfigurationException:
        """
        sync_signal = None

        executable_types = FecDataView.get_executable_types()
        if ExecutableType.USES_SIMULATION_INTERFACE in executable_types:
            sync_signal = FecDataView.get_next_sync_signal()

        # handle the sync states, but only send once if they work with
        # the simulation interface requirement
        if ExecutableType.SYNC in executable_types:
            if sync_signal == Signal.SYNC1:
                raise ConfigurationException(
                    "There can only be one SYNC signal per run. This is "
                    "because we cannot ensure the cores have not reached the "
                    "next SYNC state before we send the next SYNC. Resulting "
                    "in uncontrolled behaviour")
            sync_signal = Signal.SYNC0

        return sync_signal



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import os
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from pacman.utilities import file_format_schemas
from pacman.utilities.json_utils import placements_to_json
from jsonschema.exceptions import ValidationError
from spinn_front_end_common.data import FecDataView

PLACEMENTS_FILENAME = "placements.json"
logger = FormatAdapter(logging.getLogger(__name__))


def write_json_placements():
    """
    Runs the code to write the placements in JSON.
    """
    # Steps are tojson, validate and writefile
    progress = ProgressBar(3, "Converting to JSON Placements")

    file_path = os.path.join(
        FecDataView.get_json_dir_path(), PLACEMENTS_FILENAME)
    json_obj = placements_to_json()

    # validate the schema
    try:
        file_format_schemas.validate(json_obj, PLACEMENTS_FILENAME)
    except ValidationError as ex:
        logger.error("JSON validation exception: {}\n{}",
                     ex.message, ex.instance)

    # update and complete progress bar
    progress.update()

    # dump to json file
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(json_obj, f)

    progress.end()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.progress_bar import ProgressBar
from spinnman.model.enums import CPUState
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.scp import UpdateRuntimeProcess


def chip_runtime_updater(n_sync_steps):
    """
    Updates the runtime of an application running on a SpiNNaker machine.

    :param n_sync_steps:
    :type n_sync_steps: int or None
    """
    core_subsets = FecDataView.get_executable_types()[
        ExecutableType.USES_SIMULATION_INTERFACE]
    n_cores = len(core_subsets)
    ready_progress = ProgressBar(
        n_cores, "Waiting for cores to be either in PAUSED or READY state")
    FecDataView.get_transceiver().wait_for_cores_to_be_in_state(
        core_subsets, FecDataView.get_app_id(),
        [CPUState.PAUSED, CPUState.READY],
        error_states=frozenset({
            CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG,
            CPUState.FINISHED}), progress_bar=ready_progress, timeout=n_cores)
    ready_progress.end()

    run_until_timesteps = FecDataView.get_current_run_timesteps()
    if run_until_timesteps is None:
        infinite_run = 1
        run_until_timesteps = 0
        current_timesteps = 0
    else:
        infinite_run = 0
        current_timesteps = FecDataView.get_first_machine_time_step()

    process = UpdateRuntimeProcess(FecDataView.get_scamp_connection_selector())
    process.update_runtime(
        current_timesteps, run_until_timesteps, infinite_run, core_subsets,
        len(core_subsets), n_sync_steps)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import ExitStack
import logging
import math
from typing import Dict, Tuple
from spinn_utilities.config_holder import get_config_str_list, get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spalloc_client import Job
from spalloc_client.states import JobState
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.config_holder import get_config_int, get_config_str
from spinn_machine import Machine
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.spalloc import (
    is_server_address, SpallocClient, SpallocJob, SpallocState)
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from spinn_front_end_common.abstract_models.impl import (
    MachineAllocationController)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.utility_calls import parse_old_spalloc

logger = FormatAdapter(logging.getLogger(__name__))
_MACHINE_VERSION = 5  # Spalloc only ever works with v5 boards

#: The number of chips per board to use in calculations to ensure that
#: the number of boards allocated is enough.  This is 2 less than the maximum
#: as there are a few boards with 2 down chips in the big machine.
CALC_CHIPS_PER_BOARD = Machine.MAX_CHIPS_PER_48_BOARD - 2


class SpallocJobController(MachineAllocationController):
    __slots__ = (
        # the spalloc job object
        "_job",
        # the current job's old state
        "_state",
        "__client",
        "__closer",
        "__use_proxy"
    )

    def __init__(
            self, client: SpallocClient, job: SpallocJob,
            task: AbstractContextManager, use_proxy: bool):
        """
        :param ~spinnman.spalloc.SpallocClient client:
        :param ~spinnman.spalloc.SpallocJob job:
        :param task:
        :type task:
            ~spinn_utilities.abstract_context_manager.AbstractContextManager
        :param bool use_proxy:
        """
        if job is None:
            raise TypeError("must have a real job")
        self.__client = client
        self.__closer = task
        self._job = job
        self._state = job.get_state()
        self.__use_proxy = use_proxy
        super().__init__("SpallocJobController")

    @overrides(AbstractMachineAllocationController.extend_allocation)
    def extend_allocation(self, new_total_run_time):
        # Does Nothing in this allocator - machines are held until exit
        pass

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        super().close()
        self.__closer.close()
        self._job.destroy()
        self.__client.close()

    @overrides(AbstractMachineAllocationController.where_is_machine)
    def where_is_machine(self, chip_x, chip_y):
        """
        :param int chip_x:
        :param int chip_y:
        :rtype: tuple(int,int,int)
        """
        return self._job.where_is_machine(x=chip_x, y=chip_y)

    @overrides(MachineAllocationController._wait)
    def _wait(self):
        try:
            if self._state != SpallocState.DESTROYED:
                self._state = self._job.wait_for_state_change(self._state)
        except TypeError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            if not self._exited:
                raise e
        return self._state != SpallocState.DESTROYED

    @overrides(MachineAllocationController._teardown)
    def _teardown(self):
        if not self._exited:
            self.__closer.close()
            self._job.destroy()
            self.__client.close()
        super()._teardown()

    @overrides(AbstractMachineAllocationController.create_transceiver)
    def create_transceiver(self):
        """
        .. note::
            This allocation controller proxies the transceiver's connections
            via Spalloc. This allows it to work even outside the UNIMAN
            firewall.

        """
        if not self.__use_proxy:
            return super(SpallocJobController, self).create_transceiver()
        txrx = self._job.create_transceiver()
        txrx.ensure_board_is_ready()
        return txrx

    @overrides(AbstractMachineAllocationController.open_sdp_connection)
    def open_sdp_connection(self, chip_x, chip_y, udp_port=SCP_SCAMP_PORT):
        """
        .. note::
            This allocation controller proxies connections via Spalloc. This
            allows it to work even outside the UNIMAN firewall.

        """
        return self._job.connect_to_board(chip_x, chip_y, udp_port)

    @overrides(AbstractMachineAllocationController.open_eieio_connection)
    def open_eieio_connection(self, chip_x, chip_y):
        return self._job.open_eieio_connection(chip_x, chip_y, SCP_SCAMP_PORT)

    @overrides(AbstractMachineAllocationController.open_eieio_listener)
    def open_eieio_listener(self):
        return self._job.open_listener_connection()

    @property
    @overrides(AbstractMachineAllocationController.proxying)
    def proxying(self):
        return self.__use_proxy

    @overrides(MachineAllocationController.make_report)
    def make_report(self, filename):
        with open(filename, "w", encoding="utf-8") as report:
            report.write(f"Job: {self._job}")


class _OldSpallocJobController(MachineAllocationController):
    __slots__ = (
        # the spalloc job object
        "_job",
        # the current job's old state
        "_state"
    )

    def __init__(self, job: Job, host: str):
        """
        :param ~spalloc.job.Job job:
        """
        if job is None:
            raise TypeError("must have a real job")
        self._job = job
        self._state = job.state
        super().__init__("SpallocJobController", host)

    @overrides(AbstractMachineAllocationController.extend_allocation)
    def extend_allocation(self, new_total_run_time):
        # Does Nothing in this allocator - machines are held until exit
        pass

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        super().close()
        self._job.destroy()

    @property
    def power(self) -> bool:
        """
        :rtype: bool
        """
        return self._job.power

    def set_power(self, power: bool):
        """
        :param bool power:
        """
        self._job.set_power(power)
        if power:
            self._job.wait_until_ready()

    @overrides(AbstractMachineAllocationController.where_is_machine)
    def where_is_machine(self, chip_x, chip_y):
        return self._job.where_is_machine(chip_y=chip_y, chip_x=chip_x)

    @overrides(MachineAllocationController._wait)
    def _wait(self):
        try:
            if self._state != JobState.destroyed:
                self._state = self._job.wait_for_state_change(self._state)
        except TypeError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            if not self._exited:
                raise e
        return self._state != JobState.destroyed

    @overrides(MachineAllocationController._teardown)
    def _teardown(self):
        if not self._exited:
            self._job.close()
        super()._teardown()


_MACHINE_VERSION = 5


def spalloc_allocator(
        bearer_token: str = None) -> Tuple[
            str, int, None, bool, bool, Dict[Tuple[int, int], str], None,
            MachineAllocationController]:
    """
    Request a machine from a SPALLOC server that will fit the given
    number of chips.

    :param bearer_token: The bearer token to use
    :type bearer_token: str or None
    :return:
        host, board version, BMP details, reset on startup flag,
        auto-detect BMP flag, board address map, allocation controller
    :rtype: tuple(str, int, object, bool, bool, dict(tuple(int,int),str),
        MachineAllocationController)
    """

    spalloc_server = get_config_str("Machine", "spalloc_server")

    # Work out how many boards are needed
    if FecDataView.has_n_boards_required():
        n_boards = FecDataView.get_n_boards_required()
    else:
        n_chips = FecDataView.get_n_chips_needed()
        n_boards_float = float(n_chips) / CALC_CHIPS_PER_BOARD
        logger.info("{:.2f} Boards Required for {} chips",
                    n_boards_float, n_chips)
        # If the number of boards rounded up is less than 50% of a board
        # bigger than the actual number of boards,
        # add another board just in case.
        n_boards = int(math.ceil(n_boards_float))
        if n_boards - n_boards_float < 0.5:
            n_boards += 1

    if is_server_address(spalloc_server):
        host, connections, mac = _allocate_job_new(
            spalloc_server, n_boards, bearer_token)
    else:
        host, connections, mac = _allocate_job_old(spalloc_server, n_boards)
    return (host, _MACHINE_VERSION, None, False, False, connections, mac)


def _allocate_job_new(
        spalloc_server: str, n_boards: int,
        bearer_token: str = None) -> Tuple[
            str, Dict[Tuple[int, int], str], MachineAllocationController]:
    """
    Request a machine from an new-style spalloc server that will fit the
    given number of boards.

    :param str spalloc_server:
        The server from which the machine should be requested
    :param int n_boards: The number of boards required
    :param bearer_token: The bearer token to use
    :type bearer_token: str or None
    :rtype: tuple(str, dict(tuple(int,int),str), MachineAllocationController)
    """
    logger.info(f"Requesting job with {n_boards} boards")
    with ExitStack() as stack:
        spalloc_machine = get_config_str("Machine", "spalloc_machine")
        use_proxy = get_config_bool("Machine", "spalloc_use_proxy")
        client = SpallocClient(spalloc_server, bearer_token=bearer_token)
        stack.enter_context(client)
        job = client.create_job(n_boards, spalloc_machine)
        stack.enter_context(job)
        task = job.launch_keepalive_task()
        stack.enter_context(task)
        job.wait_until_ready()
        connections = job.get_connections()
        ProvenanceWriter().insert_board_provenance(connections)
        root = connections.get((0, 0), None)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "boards: {}",
                str(connections).replace("{", "[").replace("}", "]"))
        allocation_controller = SpallocJobController(
            client, job, task, use_proxy)
        # Success! We don't want to close the client, job or task now;
        # the allocation controller now owns them.
        stack.pop_all()
        return (root, connections, allocation_controller)


def _allocate_job_old(spalloc_server: str, n_boards: int) -> Tuple[
        str, Dict[Tuple[int, int], str], MachineAllocationController]:
    """
    Request a machine from an old-style spalloc server that will fit the
    requested number of boards.

    :param str spalloc_server:
        The server from which the machine should be requested
    :param int n_boards: The number of boards required
    :rtype: tuple(str, dict(tuple(int,int),str), MachineAllocationController)
    """
    host, port, user = parse_old_spalloc(
        spalloc_server, get_config_int("Machine", "spalloc_port"),
        get_config_str("Machine", "spalloc_user"))
    spalloc_kwargs = {
        'hostname': host,
        'port': port,
        'owner': user
    }
    spalloc_machine = get_config_str("Machine", "spalloc_machine")

    if spalloc_machine is not None:
        spalloc_kwargs['machine'] = spalloc_machine

    job, hostname, scamp_connection_data = _launch_checked_job_old(
        n_boards, spalloc_kwargs)
    machine_allocation_controller = _OldSpallocJobController(job, hostname)
    return (hostname, scamp_connection_data, machine_allocation_controller)


def _launch_checked_job_old(n_boards: int, spalloc_kwargs: dict) -> Tuple[
        Job, str, Dict[Tuple[int, int], str]]:
    """
    :rtype: tuple(~.Job, str, dict(tuple(int,int),str))
    """
    logger.info(f"Requesting job with {n_boards} boards")
    avoid_boards = get_config_str_list("Machine", "spalloc_avoid_boards")
    avoid_jobs = []
    try:
        while True:
            job = Job(n_boards, **spalloc_kwargs)
            try:
                job.wait_until_ready()
                # get param from jobs before starting, so that hanging doesn't
                # occur
                hostname = job.hostname
            except Exception as ex:
                job.destroy(str(ex))
                raise
            connections = job.connections
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("boards: {}",
                             str(connections).replace("{", "[").replace(
                                 "}", "]"))
            with ProvenanceWriter() as db:
                db.insert_board_provenance(connections)
            if hostname not in avoid_boards:
                break
            avoid_jobs.append(job)
            logger.warning(
                f"Asking for new job as {hostname} "
                f"as in the spalloc_avoid_boards list")
    finally:
        if avoid_boards:
            for key in list(connections.keys()):
                if connections[key] in avoid_boards:
                    logger.warning(
                        f"Removing connection info for {connections[key]} "
                        f"as in the spalloc avoid_boards list")
                    del connections[key]
        for avoid_job in avoid_jobs:
            avoid_job.destroy("Asked to avoid by cfg")
    return job, hostname, connections



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.messages.eieio import EIEIOType, EIEIOPrefix
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from pacman.model.resources.iptag_resource import IPtagResource

_HAS_PAYLOAD = (EIEIOType.KEY_PAYLOAD_32_BIT, EIEIOType.KEY_PAYLOAD_16_BIT)
_NO_PAYLOAD = (EIEIOType.KEY_32_BIT, EIEIOType.KEY_16_BIT)
#: Used to identify tags involved with the live packet gatherer.
TRAFFIC_IDENTIFIER = "LPG_EVENT_STREAM"


class LivePacketGatherParameters(object):
    """
    Parameter holder for :py:class:`LivePacketGather`\\ers so that they can be
    instantiated at a later date.
    """

    __slots__ = [
        '_port', '_hostname', "_tag", "_strip_sdp", "_use_prefix",
        "_key_prefix", "_prefix_type", "_message_type", "_right_shift",
        "_payload_as_time_stamps", "_use_payload_prefix", "_payload_prefix",
        "_payload_right_shift", "_n_packets_per_time_step", "_label",
        "_received_key_mask", "_translate_keys", "_translated_key_right_shift"
    ]

    def __init__(
            self, port=None, hostname=None, tag=None, strip_sdp=True,
            use_prefix=False, key_prefix=None, prefix_type=None,
            message_type=EIEIOType.KEY_32_BIT, right_shift=0,
            payload_as_time_stamps=True, use_payload_prefix=True,
            payload_prefix=None, payload_right_shift=0,
            number_of_packets_sent_per_time_step=0, label=None,
            received_key_mask=0xFFFFFFFF,
            translate_keys=False, translated_key_right_shift=0):
        """
        :raises ConfigurationException:
            If the parameters passed are known to be an invalid combination.
        """
        # pylint: disable=too-many-arguments

        # Sanity checks
        if (message_type in _HAS_PAYLOAD and use_payload_prefix and
                payload_as_time_stamps):
            raise ConfigurationException(
                "Timestamp can either be included as payload prefix or as "
                "payload to each key, not both")
        if (message_type in _NO_PAYLOAD and not use_payload_prefix and
                payload_as_time_stamps):
            raise ConfigurationException(
                "Timestamp can either be included as payload prefix or as "
                "payload to each key, but current configuration does not "
                "specify either of these")
        if (prefix_type is not None and
                not isinstance(prefix_type, EIEIOPrefix)):
            raise ConfigurationException(
                "the type of a prefix type should be of a EIEIOPrefix, "
                "which can be located in: spinnman.messages.eieio")

        self._port = port
        self._hostname = hostname
        self._tag = tag
        self._strip_sdp = strip_sdp
        self._use_prefix = use_prefix
        self._key_prefix = key_prefix
        self._prefix_type = prefix_type
        self._message_type = message_type
        self._right_shift = right_shift
        self._payload_as_time_stamps = payload_as_time_stamps
        self._use_payload_prefix = use_payload_prefix
        self._payload_prefix = payload_prefix
        self._payload_right_shift = payload_right_shift
        self._n_packets_per_time_step = number_of_packets_sent_per_time_step
        self._label = label
        self._received_key_mask = received_key_mask
        self._translate_keys = translate_keys
        self._translated_key_right_shift = translated_key_right_shift

    @property
    def port(self):
        """
        Where to send data from SpiNNaker:
        the port of the listening UDP socket.

        :rtype: int
        """
        return self._port

    @property
    def hostname(self):
        """
        Where to send data from SpiNNaker: the host name of the listening UDP
        socket.

        :rtype: bool
        """
        return self._hostname

    @property
    def tag(self):
        """
        A fixed tag ID to assign, or `None` if any tag is OK

        :rtype: int or None
        """
        return self._tag

    @property
    def strip_sdp(self):
        """
        Whether to remove SDP headers from the messages before sending.

        :rtype: bool
        """
        return self._strip_sdp

    @property
    def use_prefix(self):
        """
        Whether to use EIEIO prefix compaction on keys.

        :rtype: bool
        """
        return self._use_prefix

    @property
    def key_prefix(self):
        """
        The EIEIO key prefix to remove from messages.

        :rtype: int
        """
        return self._key_prefix

    @property
    def prefix_type(self):
        """
        The type of prefix.

        :rtype: ~spinnman.messages.eieio.EIEIOPrefix
        """
        return self._prefix_type

    @property
    def message_type(self):
        """
        The type of messages to send.

        :rtype: ~spinnman.messages.eieio.EIEIOType
        """
        return self._message_type

    @property
    def right_shift(self):
        """
        Shift to apply to keys.

        :rtype: int
        """
        return self._right_shift

    @property
    def payload_as_time_stamps(self):
        """
        Whether the payloads are timestamps.

        :rtype: bool
        """
        return self._payload_as_time_stamps

    @property
    def use_payload_prefix(self):
        """
        Whether to use prefix compaction for payloads.

        :rtype: bool
        """
        return self._use_payload_prefix

    @property
    def payload_prefix(self):
        """
        The payload prefix to remove if applying compaction.

        :rtype: int
        """
        return self._payload_prefix

    @property
    def payload_right_shift(self):
        """
        Shift to apply to payloads.

        :rtype: int
        """
        return self._payload_right_shift

    @property
    def number_of_packets_sent_per_time_step(self):
        """
        The maximum number of packets to send in a timestep.

        :rtype: int
        """
        return self._n_packets_per_time_step

    @property
    def label(self):
        """
        A label.

        :rtype: str
        """
        return self._label

    @property
    def received_key_mask(self):
        """
        A mask to select which keys are dispatched.

        :rtype: int
        """
        return self._received_key_mask

    @property
    def translate_keys(self):
        """
        Whether to apply translation to keys.

        :rtype: bool
        """
        return self._translate_keys

    @property
    def translated_key_right_shift(self):
        """
        Shift to apply in key translation.

        :rtype: int
        """
        return self._translated_key_right_shift

    def get_iptag_resource(self):
        """
        Get a description of the :py:class:`~spinn_machine.tags.IPTag`
        that the LPG for these parameters will require.

        :rtype: ~pacman.model.resources.IPtagResource
        """
        return IPtagResource(
            ip_address=self.hostname, port=self.port,
            strip_sdp=self.strip_sdp, tag=self.tag,
            traffic_identifier=TRAFFIC_IDENTIFIER)

    def __eq__(self, other):
        return (self._port == other.port and
                self._hostname == other.hostname and
                self._tag == other.tag and
                self._strip_sdp == other.strip_sdp and
                self._use_prefix == other.use_prefix and
                self._key_prefix == other.key_prefix and
                self._prefix_type == other.prefix_type and
                self._message_type == other.message_type and
                self._right_shift == other.right_shift and
                self._payload_as_time_stamps ==
                other.payload_as_time_stamps and
                self._use_payload_prefix == other.use_payload_prefix and
                self._payload_prefix == other.payload_prefix and
                self._payload_right_shift == other.payload_right_shift and
                self._n_packets_per_time_step ==
                other.number_of_packets_sent_per_time_step and
                self._label == other.label and
                self._received_key_mask == other.received_key_mask and
                self._translate_keys == other.translate_keys and
                self._translated_key_right_shift ==
                other.translated_key_right_shift)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        data = (
            self._port, self._tag, self._strip_sdp, self._use_prefix,
            self._key_prefix, self._prefix_type, self._message_type,
            self._right_shift, self._payload_as_time_stamps,
            self._use_payload_prefix, self._payload_prefix,
            self._payload_right_shift, self._n_packets_per_time_step,
            self._label, self._received_key_mask, self._translate_keys,
            self._translated_key_right_shift)
        return hash(data)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import NoOptionError
import logging
import os
import shutil
import traceback
from spinn_utilities.log import FormatAdapter
from spinn_utilities.config_holder import (
    config_options, load_config, get_config_bool, get_config_int,
    get_config_str, get_config_str_list, set_config)
from spinn_machine import Machine
from spinn_front_end_common.interface.provenance import LogStoreDB
from spinn_front_end_common.data.fec_data_writer import FecDataWriter
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))

APP_DIRNAME = 'application_generated_data_files'
FINISHED_FILENAME = "finished"
ERRORED_FILENAME = "errored"
REPORTS_DIRNAME = "reports"
TIMESTAMP_FILENAME = "time_stamp"
WARNING_LOGS_FILENAME = "warning_logs.txt"

# options names are all lower without _ inside config
_DEBUG_ENABLE_OPTS = frozenset([
    "reportsenabled",
    "clear_iobuf_during_run", "extract_iobuf"])
_REPORT_DISABLE_OPTS = frozenset([
    "clear_iobuf_during_run", "extract_iobuf"])


class ConfigHandler(object):
    """
    Superclass of AbstractSpinnakerBase that handles function only
    dependent of the configuration and the order its methods are called.
    """

    __slots__ = [

        # The writer and therefore view of the global data
        "_data_writer"

    ]

    def __init__(self, data_writer_cls=None):
        """
        :param FecDataWriter data_writer:
            The Global data writer object
        """
        load_config()

        if data_writer_cls:
            self._data_writer = data_writer_cls.setup()
        else:
            self._data_writer = FecDataWriter.setup()
        logger.set_log_store(LogStoreDB())

        # set up machine targeted data
        self._debug_configs()
        self._previous_handler()

        # Pass max_machine_cores to Machine so if effects everything!
        max_machine_core = get_config_int("Machine", "max_machine_core")
        if max_machine_core is not None:
            Machine.set_max_cores_per_chip(max_machine_core)

    def _debug_configs(self):
        """
        Adjusts and checks the configuration based on mode and
        `reports_enabled`.

        :raises ConfigurationException:
        """
        if get_config_str("Mode", "mode") == "Debug":
            for option in config_options("Reports"):
                # options names are all lower without _ inside config
                if option in _DEBUG_ENABLE_OPTS or option[:5] == "write":
                    if not get_config_bool("Reports", option):
                        set_config("Reports", option, "True")
                        logger.info("As mode == \"Debug\", [Reports] {} "
                                    "has been set to True", option)
        elif not get_config_bool("Reports", "reportsEnabled"):
            for option in config_options("Reports"):
                # options names are all lower without _ inside config
                if option in _REPORT_DISABLE_OPTS or option[:5] == "write":
                    if not get_config_bool("Reports", option):
                        set_config("Reports", option, "False")
                        logger.info(
                            "As reportsEnabled == \"False\", [Reports] {} "
                            "has been set to False", option)
        if get_config_bool("Machine", "virtual_board"):
            # TODO handle in the execute methods
            if get_config_bool("Reports", "write_energy_report"):
                set_config("Reports", "write_energy_report", "False")
                logger.info("[Reports]write_energy_report has been set to "
                            "False as using virtual boards")

    def _previous_handler(self):
        self._error_on_previous("loading_algorithms")
        self._error_on_previous("application_to_machine_graph_algorithms")
        self._error_on_previous("machine_graph_to_machine_algorithms")
        self._error_on_previous("machine_graph_to_virtual_machine_algorithms")

    def _error_on_previous(self, option):
        try:
            get_config_str_list("Mapping", option)
        except NoOptionError:
            # GOOD!
            return
        raise ConfigurationException(
            f"cfg setting {option} is no longer supported! "
            "See https://spinnakermanchester.github.io/common_pages/"
            "Algorithms.html.")

    def _adjust_config(self, runtime):
        """
        Adjust and checks the configuration based on runtime

        :param runtime:
        :type runtime: int or bool
        :raises ConfigurationException:
        """
        if runtime is None:
            if get_config_bool("Reports", "write_energy_report"):
                set_config("Reports", "write_energy_report", "False")
                logger.info("[Reports]write_energy_report has been set to "
                            "False as runtime is set to forever")

    def _remove_excess_folders(
            self, max_kept, starting_directory, remove_errored_folders):
        try:
            files_in_report_folder = os.listdir(starting_directory)

            # while there's more than the valid max, remove the oldest one
            if len(files_in_report_folder) > max_kept:

                # sort files into time frame
                files_in_report_folder.sort(
                    key=lambda temp_file: os.path.getmtime(
                        os.path.join(starting_directory, temp_file)))

                # remove only the number of files required, and only if they
                # have the finished flag file created
                num_files_to_remove = len(files_in_report_folder) - max_kept
                files_removed = 0
                files_not_closed = 0
                for current_oldest_file in files_in_report_folder:
                    finished_flag = os.path.join(os.path.join(
                        starting_directory, current_oldest_file),
                        FINISHED_FILENAME)
                    errored_flag = os.path.join(os.path.join(
                        starting_directory, current_oldest_file),
                        ERRORED_FILENAME)
                    finished_flag_exists = os.path.exists(finished_flag)
                    errored_flag_exists = os.path.exists(errored_flag)
                    if finished_flag_exists and (
                            not errored_flag_exists or remove_errored_folders):
                        shutil.rmtree(os.path.join(
                            starting_directory, current_oldest_file),
                            ignore_errors=True)
                        files_removed += 1
                    else:
                        files_not_closed += 1
                    if files_removed + files_not_closed >= num_files_to_remove:
                        break
                if files_not_closed > max_kept // 4:
                    logger.warning(
                        "{} has {} old reports that have not been closed",
                        starting_directory, files_not_closed)
        except IOError:
            # This might happen if there is an open file, or more than one
            # process in the same folder, but we shouldn't die because of it
            pass

    def _set_up_report_specifics(self):
        # clear and clean out folders considered not useful anymore
        report_dir_path = self._data_writer.get_report_dir_path()
        if os.listdir(report_dir_path):
            self._remove_excess_folders(
                get_config_int("Reports", "max_reports_kept"),
                report_dir_path,
                get_config_bool("Reports", "remove_errored_folders"))

        # store timestamp in latest/time_stamp for provenance reasons
        timestamp_dir_path = self._data_writer.get_timestamp_dir_path()
        time_of_run_file_name = os.path.join(
            timestamp_dir_path, TIMESTAMP_FILENAME)
        _, timestamp = os.path.split(timestamp_dir_path)
        with open(time_of_run_file_name, "w", encoding="utf-8") as f:
            f.writelines(timestamp)
            f.write("\n")
            f.write("Traceback of setup call:\n")
            traceback.print_stack(file=f)

    def __write_named_file(self, file_name):
        app_file_name = os.path.join(
            self._data_writer.get_timestamp_dir_path(), file_name)
        with open(app_file_name, "w", encoding="utf-8") as f:
            f.writelines("file_name")

    def write_finished_file(self):
        """
        Write a finished file that allows file removal to only remove
        folders that are finished.
        """
        self.__write_named_file(FINISHED_FILENAME)

    def write_errored_file(self):
        """
        Writes an ``errored`` file that allows file removal to only remove
        folders that have errors if requested to do so
        """
        self.__write_named_file(ERRORED_FILENAME)



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from collections import defaultdict
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import (
    ProvenanceReader, ProvenanceWriter)
from .bit_field_summary import BitFieldSummary
from spinn_front_end_common.utilities.utility_objs import ExecutableType

logger = FormatAdapter(logging.getLogger(__name__))
_FILE_NAME = "bit_field_compressed_summary.rpt"
# provenance data item names

MERGED_NAME = "bit_fields_merged"
NOT_APPLICABLE = "N/A"


def generate_provenance_item(x, y, bit_fields_merged):
    """
    Generates a provenance item in the format BitFieldCompressorReport expects.

    :param x:
    :param y:
    :param bit_fields_merged:
    """
    with ProvenanceWriter() as db:
        db.insert_router(x, y, MERGED_NAME, bit_fields_merged)


def bitfield_compressor_report():
    """
    Generates a report that shows the impact of the compression of
    bitfields into the routing table.

    :return: a summary, or `None` if the report file can't be written
    :rtype: BitFieldSummary
    """
    file_name = os.path.join(FecDataView.get_run_dir_path(), _FILE_NAME)
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            _write_report(f)
    except IOError:
        logger.exception("Generate_placement_reports: Can't open file"
                         " {} for writing.", _FILE_NAME)


def _merged_component(to_merge_per_chip, writer):
    """
    Report how many bitfields were merged into the router.

    :param dict([int, int], int: to_merge_per_chip: number of bitfields
        that could be merged per chip
    :param ~io.FileIO writer: file writer.
    :return: tuple containing 4 elements.
     1. min_bit_fields merged in a chip,
     2. the max bit_fields merged in a chip,
     3. the total bit_fields_merged into all the routers.
     4. average number of bit-fields merged on the routers.
    :rtype: tuple(int or str, int or str, int or str, float or str)
    """
    top_bit_field = 0
    min_bit_field = sys.maxsize
    total_bit_fields_merged = 0
    average_per_chip_merged = 0
    n_chips = 0
    to_merge_chips = set(to_merge_per_chip.keys())

    found = False
    with ProvenanceReader() as db:
        for (x, y, merged) in db.get_router_by_chip(
                MERGED_NAME):
            if (x, y) not in to_merge_per_chip:
                continue
            to_merge = to_merge_per_chip[x, y]
            to_merge_chips.discard((x, y))
            found = True
            writer.write(
                f"Chip {x}:{y} has {merged} bitfields out of {to_merge} "
                f"merged into it. Which is {merged / to_merge:.2%}\n")
            total_bit_fields_merged += int(merged)
            if merged > top_bit_field:
                top_bit_field = merged
            if merged < min_bit_field:
                min_bit_field = merged
            average_per_chip_merged += merged
            n_chips += 1

    if found:
        average_per_chip_merged = (
            float(average_per_chip_merged) / float(n_chips))
    else:
        min_bit_field = NOT_APPLICABLE
        top_bit_field = NOT_APPLICABLE
        total_bit_fields_merged = NOT_APPLICABLE
        average_per_chip_merged = NOT_APPLICABLE

    if len(to_merge_chips) > 0:
        writer.write(
            f"The Chips {to_merge_chips} had bitfields.\n"
            "But no record was found of any attempt to merge them.\n")

    return (min_bit_field, top_bit_field, total_bit_fields_merged,
            average_per_chip_merged)


def _compute_to_merge_per_chip():
    """
    :rtype: tuple(int, int, int, float or int)
    """
    total_to_merge = 0
    to_merge_per_chip = defaultdict(int)

    for partition in FecDataView.iterate_partitions():
        for edge in partition.edges:
            splitter = edge.post_vertex.splitter
            for vertex in splitter.get_source_specific_in_coming_vertices(
                    partition.pre_vertex, partition.identifier):
                if not isinstance(vertex, AbstractHasAssociatedBinary):
                    continue
                if vertex.get_binary_start_type() == ExecutableType.SYSTEM:
                    continue
                place = FecDataView.get_placement_of_vertex(vertex)
                to_merge_per_chip[place.chip] += 1
                total_to_merge += 1

    return total_to_merge, to_merge_per_chip


def _before_merge_component(total_to_merge, to_merge_per_chip):
    """
    :rtype: tuple(int, int, int, float or int)
    """
    max_bit_fields_on_chip = 0
    min_bit_fields_on_chip = sys.maxsize

    for bitfield_count in to_merge_per_chip.values():
        if bitfield_count > max_bit_fields_on_chip:
            max_bit_fields_on_chip = bitfield_count
        if bitfield_count < min_bit_fields_on_chip:
            min_bit_fields_on_chip = bitfield_count

    if len(to_merge_per_chip) == 0:
        average = 0
    else:
        average = float(total_to_merge) / float(len(to_merge_per_chip))

    return max_bit_fields_on_chip, min_bit_fields_on_chip, average


def _write_report(writer):
    """
    Writes the report.

    :param ~io.FileIO writer: the file writer
    :return: a summary
    :rtype: BitFieldSummary
    """
    total_to_merge, to_merge_per_chip = _compute_to_merge_per_chip()
    (max_to_merge_per_chip, low_to_merge_per_chip,
     average_per_chip_to_merge) = _before_merge_component(
        total_to_merge, to_merge_per_chip)
    (min_bit_field, top_bit_field, total_bit_fields_merged,
     average_per_chip_merged) = _merged_component(to_merge_per_chip, writer)
    writer.write(
        f"\n\nBefore merge there where {total_to_merge} bitfields on "
        f"{len(to_merge_per_chip)} Chips ranging from {max_to_merge_per_chip} "
        f"to {low_to_merge_per_chip} bitfields per chip with an average "
        f"of {average_per_chip_to_merge}")
    writer.write(
        f"\nSuccessfully merged {total_bit_fields_merged} bitfields ranging "
        f"from {top_bit_field} to {min_bit_field} bitfields per chip with an "
        f"average of {average_per_chip_merged}")
    if total_to_merge:
        if total_bit_fields_merged == NOT_APPLICABLE:
            writer.write(f"\nNone of the {total_to_merge} bitfields merged")
        else:
            writer.write("\nIn total {:.2%} of the bitfields merged".format(
                total_bit_fields_merged / total_to_merge))

    return BitFieldSummary(
        lowest_per_chip=min_bit_field, max_per_chip=top_bit_field,
        total_merged=total_bit_fields_merged,
        total_to_merge=total_to_merge,
        max_to_merge_per_chip=max_to_merge_per_chip,
        low_to_merge_per_chip=low_to_merge_per_chip,
        average_per_chip_merged=average_per_chip_merged,
        average_per_chip_to_merge=average_per_chip_to_merge)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from spinn_utilities.progress_bar import ProgressBar
from pacman.utilities import file_format_schemas
from pacman.model.routing_tables.multicast_routing_tables import to_json
from spinn_front_end_common.data import FecDataView

ROUTING_TABLES_FILENAME = "routing_tables.json"


def write_json_routing_tables(router_tables):
    """
    Runs the code to write the machine in Java readable JSON.

    :param ~pacman.model.routing_tables.MulticastRoutingTables router_tables:
        Routing Tables to convert. Could be uncompressed or compressed
    :param str json_folder: the folder to which the JSON are being written
    """
    # Steps are tojson, validate and writefile
    progress = ProgressBar(3, "Converting to JSON RouterTables")

    file_path = os.path.join(
        FecDataView.get_json_dir_path(), ROUTING_TABLES_FILENAME)
    json_obj = to_json(router_tables)

    if progress:
        progress.update()

    # validate the schema
    file_format_schemas.validate(json_obj, ROUTING_TABLES_FILENAME)

    # update and complete progress bar
    if progress:
        progress.update()

    # dump to json file
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(json_obj, f)

    if progress:
        progress.end()

    return file_path



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from spinnman.constants import MAX_TAG_ID
from spinn_front_end_common.data import FecDataView


def tags_loader():
    """
    Loads tags onto the machine.
    """
    # clear all the tags from the Ethernet connection, as nothing should
    # be allowed to use it (no two apps should use the same Ethernet
    # connection at the same time)
    transceiver = FecDataView.get_transceiver()
    progress = ProgressBar(MAX_TAG_ID, "Clearing tags")
    for tag_id in progress.over(range(MAX_TAG_ID)):
        transceiver.clear_ip_tag(tag_id)

    # Use tags object to supply tag info if it is supplied
    tags = FecDataView.get_tags()
    iptags = list(tags.ip_tags)
    reverse_iptags = list(tags.reverse_ip_tags)

    # Load the IP tags and the Reverse IP tags
    progress = ProgressBar(
        len(iptags) + len(reverse_iptags), "Loading Tags")
    for ip_tag in progress.over(iptags, False):
        transceiver.set_ip_tag(ip_tag)
    for reverse_ip_tag in progress.over(reverse_iptags, False):
        transceiver.set_reverse_ip_tag(reverse_ip_tag)
    progress.end()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView
from spinn_machine import Machine, Router

AREA_CODE_REPORT_NAME = "board_chip_report.txt"


def board_chip_report():
    """
    Creates a report that states where in SDRAM each region is.
    """
    machine = FecDataView.get_machine()
    # create file path
    directory_name = os.path.join(
        FecDataView.get_run_dir_path(), AREA_CODE_REPORT_NAME)
    # create the progress bar for end users
    progress_bar = ProgressBar(
        len(machine.ethernet_connected_chips),
        "Writing the board chip report")

    # iterate over ethernet chips and then the chips on that board
    with open(directory_name, "w", encoding="utf-8") as writer:
        _write_report(writer, machine, progress_bar)


def _write_report(writer, machine, progress_bar):
    """
    :param ~io.FileIO writer:
    :param ~spinn_machine.Machine machine:
    :param ~spinn_utilities.progress_bar.ProgressBar progress_bar:
    """
    down_links = list()
    down_chips = list()
    down_cores = list()
    for e_chip in progress_bar.over(machine.ethernet_connected_chips):
        existing_chips = list()
        for l_x, l_y in machine.local_xys:
            x, y = machine.get_global_xy(l_x, l_y, e_chip.x, e_chip.y)
            if machine.is_chip_at(x, y):
                chip = machine.get_chip_at(x, y)
                existing_chips.append(
                    f"({x}, {y}, P: {chip.get_physical_core_id(0)})")
                down_procs = set(range(Machine.DEFAULT_MAX_CORES_PER_CHIP))
                for proc in chip.processors:
                    down_procs.remove(proc.processor_id)
                for p in down_procs:
                    phys_p = chip.get_physical_core_id(p)
                    core = p
                    if phys_p is not None:
                        core = -phys_p
                    down_cores.append((l_x, l_y, core, e_chip.ip_address))
            else:
                down_chips.append((l_x, l_y, e_chip.ip_address))
            for link in range(Router.MAX_LINKS_PER_ROUTER):
                if not machine.is_link_at(x, y, link):
                    down_links.append((l_x, l_y, link, e_chip.ip_address))

        existing_chips = ", ".join(existing_chips)
        writer.write(
            f"board with IP address: {e_chip.ip_address} has chips"
            f" {existing_chips}\n")

    down_chips_out = ":".join(
        f"{x},{y},{ip}" for x, y, ip in down_chips)
    down_cores_out = ":".join(
        f"{x},{y},{p},{ip}" for x, y, p, ip in down_cores)
    down_links_out = ":".join(
        f"{x},{y},{l},{ip}" for x, y, l, ip in down_links)
    writer.write(f"Down chips: {down_chips_out}\n")
    writer.write(f"Down cores: {down_cores_out}\n")
    writer.write(f"Down Links: {down_links_out}\n")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io


class DataRowWriter(io.RawIOBase):
    __slots__ = [
        "_x",
        "_y",
        "_p",
        "_targets",
        "_data",
    ]

    def __init__(self, x, y, p, targets):
        super().__init__()
        self._x = x
        self._y = y
        self._p = p
        self._targets = targets
        self._data = bytearray()

    def write(self, data):
        assert self.closed is False
        self._data += data

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return False

    def truncate(self, size=None):
        # Ignore
        pass

    def fileno(self):
        raise OSError

    def close(self):
        """
        Closes the writer if not already closed.
        """
        if not self.closed:
            self._targets.write_data_spec(
                self._x, self._y, self._p, self._data)
        super().close()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_front_end_common.data import FecDataView

_REPORT_FILENAME = "tags_on_machine.txt"


def tags_from_machine_report():
    """
    Describes what the tags actually present on the machine are.
    """
    filename = os.path.join(FecDataView.get_run_dir_path(), _REPORT_FILENAME)
    tags = _get_tags()
    with open(filename, "w", encoding="utf-8") as f:
        f.write("Tags actually read off the machine\n")
        f.write("==================================\n")
        for tag in tags:
            f.write(f"{repr(tag)}\n")


def _get_tags():
    try:
        return FecDataView.get_transceiver().get_tags()
    except Exception as e:  # pylint: disable=broad-except
        return [e]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.scp import ClearIOBUFProcess
from spinn_front_end_common.utilities.utility_objs import ExecutableType


def chip_io_buf_clearer():
    """
    Clears the logging output buffer of an application running on a
    SpiNNaker machine.
    """
    executable_types = FecDataView.get_executable_types()
    if ExecutableType.USES_SIMULATION_INTERFACE in executable_types:
        core_subsets = \
            executable_types[ExecutableType.USES_SIMULATION_INTERFACE]

        process = ClearIOBUFProcess(
            FecDataView.get_scamp_connection_selector())
        process.clear_iobuf(core_subsets, len(core_subsets))



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import ctypes
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.ordered_set import OrderedSet
from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.eieio.command_messages import EventStopRequest
from spinnman.messages.eieio import EIEIOType
from spinnman.messages.eieio.data_messages import EIEIODataMessage
from data_specification.constants import BYTES_PER_WORD
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import (
    BufferableRegionTooSmall, SpinnFrontEndException)
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement, locate_extra_monitor_mc_receiver)
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import (BuffersSentDeque, BufferDatabase)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost, AbstractSendsBuffersFromHost)
from spinn_front_end_common.utility_models.streaming_context_manager import (
    StreamingContextManager)
from .recording_utilities import get_recording_header_size


logger = FormatAdapter(logging.getLogger(__name__))

# The minimum size of any message - this is the headers plus one entry
_MIN_MESSAGE_SIZE = EIEIODataMessage.min_packet_length(
    eieio_type=EIEIOType.KEY_32_BIT, is_timestamp=True)

# The number of bytes in each key to be sent
_N_BYTES_PER_KEY = EIEIOType.KEY_32_BIT.key_bytes  # @UndefinedVariable

_SDP_MAX_PACKAGE_SIZE = 272

TRAFFIC_IDENTIFIER = "BufferTraffic"

VERIFY = False


class _RecordingRegion(ctypes.LittleEndianStructure):
    """
    Recording Region data
    """
    _fields_ = [
        # Space available for recording
        ("space", ctypes.c_uint32),
        # The size of the recording region
        ("size", ctypes.c_uint32, 31),
        # Whether any data is missing
        ("missing", ctypes.c_uint32, 1),
        # The address of the data
        ("data", ctypes.c_uint32)
    ]


class BufferManager(object):
    """
    Manager of send buffers.
    """

    __slots__ = [

        # Set of vertices with buffers to be sent
        "_sender_vertices",

        # Dictionary of sender vertex -> buffers sent
        "_sent_messages",

        # Support class to help call Java
        "_java_caller",

        # The machine controller, in case it wants to make proxied connections
        # for us
        "_machine_controller"
    ]

    def __init__(self):

        # Set of vertices with buffers to be sent
        self._sender_vertices = set()

        # Dictionary of sender vertex -> buffers sent
        self._sent_messages = dict()

        if FecDataView.has_java_caller():
            with BufferDatabase() as db:
                db.write_session_credentials_to_db()
            self._java_caller = FecDataView.get_java_caller()
            if get_config_bool("Machine", "enable_advanced_monitor_support"):
                self._java_caller.set_advanced_monitors()
        else:
            self._java_caller = None

        for placement in FecDataView.iterate_placements_by_vertex_type(
                AbstractSendsBuffersFromHost):
            vertex = placement.vertex
            if vertex.buffering_input():
                self._sender_vertices.add(vertex)

    def _request_data(self, placement_x, placement_y, address, length):
        """
        Uses the extra monitor cores for data extraction.

        :param int placement_x:
            the placement X coordinate where data is to be extracted from
        :param int placement_y:
            the placement Y coordinate where data is to be extracted from
        :param int address: the memory address to start at
        :param int length: the number of bytes to extract
        :return: data as a byte array
        :rtype: bytearray
        """
        # pylint: disable=too-many-arguments
        if not get_config_bool("Machine", "enable_advanced_monitor_support"):
            return FecDataView.read_memory(
                placement_x, placement_y, address, length)

        # Round to word boundaries
        initial = address % BYTES_PER_WORD
        address -= initial
        length += initial
        final = (BYTES_PER_WORD - (length % BYTES_PER_WORD)) % BYTES_PER_WORD
        length += final

        sender = FecDataView.get_monitor_by_xy(placement_x, placement_y)
        receiver = locate_extra_monitor_mc_receiver(placement_x, placement_y)
        extra_mon_data = receiver.get_data(
            sender, FecDataView.get_placement_of_vertex(sender),
            address, length)
        if VERIFY:
            txrx_data = FecDataView.read_memory(
                placement_x, placement_y, address, length)
            self._verify_data(extra_mon_data, txrx_data)

        # If we rounded to word boundaries, strip the padding junk
        if initial and final:
            return extra_mon_data[initial:-final]
        elif initial:
            return extra_mon_data[initial:]
        elif final:
            return extra_mon_data[:-final]
        else:
            return extra_mon_data

    def _verify_data(self, extra_mon_data, txrx_data):
        for index, (extra_mon_element, txrx_element) in enumerate(
                zip(extra_mon_data, txrx_data)):
            if extra_mon_element != txrx_element:
                raise ValueError(f"WRONG (at index {index})")

    def load_initial_buffers(self):
        """
        Load the initial buffers for the senders using memory writes.
        """
        total_data = 0
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                total_data += vertex.get_region_buffer_size(region)

        progress = ProgressBar(total_data, "Loading buffers")
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                self._send_initial_messages(vertex, region, progress)
        progress.end()

    def reset(self):
        """
        Resets the buffered regions to start transmitting from the beginning
        of its expected regions and clears the buffered out data files.
        """
        with BufferDatabase() as db:
            db.write_session_credentials_to_db()

        # rewind buffered in
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                vertex.rewind(region)

    def resume(self):
        """
        Resets any data structures needed before starting running again.
        """

    def clear_recorded_data(self, x, y, p, recording_region_id):
        """
        Removes the recorded data stored in memory.

        :param int x: placement X coordinate
        :param int y: placement Y coordinate
        :param int p: placement processor ID
        :param int recording_region_id: the recording region ID
        """
        with BufferDatabase() as db:
            db.clear_region(x, y, p, recording_region_id)

    def _create_message_to_send(self, size, vertex, region):
        """
        Creates a single message to send with the given boundaries.

        :param int size: The number of bytes available for the whole packet
        :param AbstractSendsBuffersFromHost vertex:
            The vertex to get the keys from
        :param int region: The region of the vertex to get keys from
        :return: A new message, or `None` if no keys can be added
        :rtype: None or ~spinnman.messages.eieio.data_messages.EIEIODataMessage
        """

        # If there are no more messages to send, return None
        if not vertex.is_next_timestamp(region):
            return None

        # Create a new message
        next_timestamp = vertex.get_next_timestamp(region)
        message = EIEIODataMessage.create(
            EIEIOType.KEY_32_BIT, timestamp=next_timestamp)

        # If there is no room for the message, return None
        if message.size + _N_BYTES_PER_KEY > size:
            return None

        # Add keys up to the limit
        bytes_to_go = size - message.size
        while (bytes_to_go >= _N_BYTES_PER_KEY and
                vertex.is_next_key(region, next_timestamp)):

            key = vertex.get_next_key(region)
            message.add_key(key)
            bytes_to_go -= _N_BYTES_PER_KEY

        return message

    def _send_initial_messages(self, vertex, region, progress):
        """
        Send the initial set of messages.

        :param AbstractSendsBuffersFromHost vertex:
            The vertex to get the keys from
        :param int region: The region to get the keys from
        :return: A list of messages
        :rtype: list(~spinnman.messages.eieio.data_messages.EIEIODataMessage)
        """

        # Get the vertex load details
        # region_base_address = self._locate_region_address(region, vertex)
        placement = FecDataView.get_placement_of_vertex(vertex)
        region_base_address = locate_memory_region_for_placement(
            placement, region)

        # Add packets until out of space
        sent_message = False
        bytes_to_go = vertex.get_region_buffer_size(region)
        if bytes_to_go % 2 != 0:
            raise SpinnFrontEndException(
                f"The buffer region of {vertex} must be divisible by 2")
        all_data = b""
        if vertex.is_empty(region):
            sent_message = True
        else:
            min_size_of_packet = _MIN_MESSAGE_SIZE
            while (vertex.is_next_timestamp(region) and
                    bytes_to_go > min_size_of_packet):
                space_available = min(bytes_to_go, _SDP_MAX_PACKAGE_SIZE)
                next_message = self._create_message_to_send(
                    space_available, vertex, region)
                if next_message is None:
                    break

                # Write the message to the memory
                data = next_message.bytestring
                all_data += data
                sent_message = True

                # Update the positions
                bytes_to_go -= len(data)
                progress.update(len(data))

        if not sent_message:
            raise BufferableRegionTooSmall(
                f"The buffer size {bytes_to_go} is too small for any data to "
                f"be added for region {region} of vertex {vertex}")

        # If there are no more messages and there is space, add a stop request
        if (not vertex.is_next_timestamp(region) and
                bytes_to_go >= EventStopRequest.get_min_packet_length()):
            data = EventStopRequest().bytestring
            # logger.debug(
            #    "Writing stop message of {} bytes to {} on {}, {}, {}"
            #         len(data), hex(region_base_address),
            #         placement.x, placement.y, placement.p)
            all_data += data
            bytes_to_go -= len(data)
            progress.update(len(data))
            self._sent_messages[vertex] = BuffersSentDeque(
                region, sent_stop_message=True)

        # Do the writing all at once for efficiency
        FecDataView.write_memory(
            placement.x, placement.y, region_base_address, all_data)

    def __get_recording_placements(self):
        """
        :rtype: list(~.Placement)
        """
        recording_placements = list()
        for placement in FecDataView.iterate_placements_by_vertex_type(
                AbstractReceiveBuffersToHost):
            recording_placements.append(placement)
        return recording_placements

    def get_placement_data(self):
        if self._java_caller is not None:
            self.__get_data_for_placements_using_java()
        else:
            recording_placements = self.__get_recording_placements()
            if get_config_bool(
                    "Machine", "enable_advanced_monitor_support"):
                self.__python_get_data_for_placements_with_monitors(
                    recording_placements)
            else:
                self.__python_get_data_for_placements(recording_placements)

    def __get_data_for_placements_using_java(self):
        logger.info("Starting buffer extraction using Java")
        self._java_caller.set_placements(
            FecDataView.iterate_placements_by_vertex_type(
                AbstractReceiveBuffersToHost))
        self._java_caller.get_all_data()

    def __python_get_data_for_placements_with_monitors(
            self, recording_placements):
        """
        :param ~pacman.model.placements.Placements recording_placements:
            Where to get the data from.
        """
        # locate receivers
        receivers = list(OrderedSet(
            locate_extra_monitor_mc_receiver(placement.x, placement.y)
            for placement in recording_placements))

        # update transaction id from the machine for all extra monitors
        for extra_mon in FecDataView.iterate_monitors():
            extra_mon.update_transaction_id_from_machine()

        with StreamingContextManager(receivers):
            # get data
            self.__python_get_data_for_placements(recording_placements)

    def __python_get_data_for_placements(self, recording_placements):
        """
        :param ~pacman.model.placements.Placements recording_placements:
            Where to get the data from.
        """
        # get data
        progress = ProgressBar(
            len(recording_placements),
            "Extracting buffers from the last run")

        with BufferDatabase() as db:
            for placement in progress.over(recording_placements):
                self._retreive_by_placement(db, placement)

    def get_data_by_placement(self, placement, recording_region_id):
        """
        Get the data container for all the data retrieved
        during the simulation from a specific region area of a core.

        :param ~pacman.model.placements.Placement placement:
            the placement to get the data from
        :param int recording_region_id: desired recording data region
        :return: an array contained all the data received during the
            simulation, and a flag indicating if any data was missing
        :rtype: tuple(bytearray, bool)
        """
        # Ensure that any transfers in progress are complete first
        if not isinstance(placement.vertex, AbstractReceiveBuffersToHost):
            raise NotImplementedError(
                f"vertex {placement.vertex} does not implement "
                "AbstractReceiveBuffersToHost so no data read")

        # data flush has been completed - return appropriate data
        with BufferDatabase() as db:
            return db.get_region_data(
                placement.x, placement.y, placement.p, recording_region_id)

    def _retreive_by_placement(self, db, placement):
        """
        Retrieve the data for a vertex; must be locked first.

        :param BufferDatabase db: database to store into
        :param ~pacman.model.placements.Placement placement:
            the placement to get the data from
        :param int recording_region_id: desired recording data region
        """
        vertex = placement.vertex
        addr = vertex.get_recording_region_base_address(placement)
        sizes_and_addresses = self._get_region_information(
                addr, placement.x, placement.y)

        # Read the data if not already received
        for region in vertex.get_recorded_region_ids():
            # Now read the data and store it
            size, addr, missing = sizes_and_addresses[region]
            data = self._request_data(
                placement.x, placement.y, addr, size)
            db.store_data_in_region_buffer(
                placement.x, placement.y, placement.p, region, missing, data)

    def _get_region_information(self, addr, x, y):
        """
        Get the recording information from all regions of a core.

        :param addr: The recording region base address
        :param x: The X coordinate of the chip containing the data
        :param y: The Y coordinate of the chip containing the data
        """
        transceiver = FecDataView.get_transceiver()
        n_regions = transceiver.read_word(x, y, addr)
        n_bytes = get_recording_header_size(n_regions)
        data = transceiver.read_memory(
            x, y, addr + BYTES_PER_WORD, n_bytes - BYTES_PER_WORD)
        data_type = _RecordingRegion * n_regions
        regions = data_type.from_buffer_copy(data)
        sizes_and_addresses = [
            (r.size, r.data, bool(r.missing)) for r in regions]
        return sizes_and_addresses

    @property
    def sender_vertices(self):
        """
        The vertices which are buffered.

        :rtype: iterable(AbstractSendsBuffersFromHost)
        """
        return self._sender_vertices



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)


class AbstractCanReset(object, metaclass=AbstractBase):
    """
    Indicates an object that can be reset to time 0.

    This is used when AbstractSpinnakerBase.reset is called.
    All Vertices and all edges in the original graph
    (the one added to by the user) will be checked and reset.
    """
    __slots__ = []

    @abstractmethod
    def reset_to_first_timestep(self):
        """
        Reset the object to first time step.
        """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
import time
from spinnman.spalloc.spalloc_job import SpallocJob
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.base_database import BaseDatabase

_SECONDS_TO_MICRO_SECONDS_CONVERSION = 1000
#: Name of the database in the data folder


def _timestamp():
    return int(time.time() * _SECONDS_TO_MICRO_SECONDS_CONVERSION)


class BufferDatabase(BaseDatabase):
    """
    Specific implementation of the Database for SQLite 3.

    There should only ever be a single Database Object in use at any time.
    In the case of application_graph_changed the first should closed and
    a new one created.

    If 2 database objects where opened with the database_file they hold the
    same data. Unless someone else deletes that file.

    .. note::
        *Not thread safe on the same database file!*
        Threads can access different DBs just fine.
    """

    __slots__ = []

    def clear_region(self, x, y, p, region):
        """
        Clears the data for a single region.

        .. note::
            This method *loses information!*

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data to be cleared
        :return: True if any region was changed
        :rtype: bool
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT region_id FROM region_view
                    WHERE x = ? AND y = ? AND processor = ?
                        AND local_region_index = ? AND fetches > 0 LIMIT 1
                    """, (x, y, p, region)):
                locus = (row["region_id"], )
                break
            else:
                return False
            cursor.execute(
                """
                UPDATE region SET
                    content = CAST('' AS BLOB), content_len = 0,
                    fetches = 0, append_time = NULL
                WHERE region_id = ?
                """, locus)
            cursor.execute(
                """
                DELETE FROM region_extra WHERE region_id = ?
                """, locus)
            return True

    def _read_contents(self, cursor, region_id):
        """
        :param ~sqlite3.Cursor cursor:
        :param int region_id:
        :rtype: memoryview
        """
        for row in cursor.execute(
                """
                SELECT content
                FROM region_view
                WHERE region_id = ?
                LIMIT 1
                """, (region_id,)):
            data = row["content"]
            break
        else:
            raise LookupError(f"no record for region {region_id}")

        c_buffer = None
        for row in cursor.execute(
                """
                SELECT r.content_len + (
                    SELECT SUM(x.content_len)
                    FROM region_extra AS x
                    WHERE x.region_id = r.region_id) AS len
                FROM region AS r WHERE region_id = ? LIMIT 1
                """, (region_id, )):
            if row["len"] is not None:
                c_buffer = bytearray(row["len"])
                c_buffer[:len(data)] = data

        if c_buffer is not None:
            idx = len(data)
            for row in cursor.execute(
                    """
                    SELECT content FROM region_extra
                    WHERE region_id = ? ORDER BY extra_id ASC
                    """, (region_id, )):
                item = row["content"]
                c_buffer[idx:idx + len(item)] = item
                idx += len(item)
            data = c_buffer
        return memoryview(data)

    def _get_region_id(self, cursor, x, y, p, region):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :param int region:
        """
        for row in cursor.execute(
                """
                SELECT region_id FROM region_view
                WHERE x = ? AND y = ? AND processor = ?
                    AND local_region_index = ?
                LIMIT 1
                """, (x, y, p, region)):
            return row["region_id"]
        core_id = self._get_core_id(cursor, x, y, p)
        cursor.execute(
            """
            INSERT INTO region(
                core_id, local_region_index, content, content_len, fetches)
            VALUES(?, ?, CAST('' AS BLOB), 0, 0)
            """, (core_id, region))
        return cursor.lastrowid

    def store_data_in_region_buffer(self, x, y, p, region, missing, data):
        """
        Store some information in the corresponding buffer for a
        specific chip, core and recording region.

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data to be stored
        :param bool missing: Whether any data is missing
        :param bytearray data: data to be stored

            .. note::
                    Must be shorter than 1GB
        """

        # pylint: disable=too-many-arguments, unused-argument
        # TODO: Use missing
        datablob = sqlite3.Binary(data)
        with self.transaction() as cursor:
            region_id = self._get_region_id(cursor, x, y, p, region)
            if self.__use_main_table(cursor, region_id):
                cursor.execute(
                    """
                    UPDATE region SET
                        content = CAST(? AS BLOB),
                        content_len = ?,
                        fetches = fetches + 1,
                        append_time = ?
                    WHERE region_id = ?
                    """, (datablob, len(data), _timestamp(), region_id))
            else:
                cursor.execute(
                    """
                    UPDATE region SET
                        fetches = fetches + 1,
                        append_time = ?
                    WHERE region_id = ?
                    """, (_timestamp(), region_id))
                assert cursor.rowcount == 1
                cursor.execute(
                    """
                    INSERT INTO region_extra(
                        region_id, content, content_len)
                    VALUES (?, CAST(? AS BLOB), ?)
                    """, (region_id, datablob, len(data)))
            assert cursor.rowcount == 1

    def __use_main_table(self, cursor, region_id):
        """
        :param ~sqlite3.Cursor cursor:
        :param int region_id:
        """
        for row in cursor.execute(
                """
                SELECT COUNT(*) AS existing FROM region
                WHERE region_id = ? AND fetches = 0
                LIMIT 1
                """, (region_id, )):
            existing = row["existing"]
            return existing == 1
        return False

    def get_region_data(self, x, y, p, region):
        """
        Get the data stored for a given region of a given core.

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data
        :return:
            A buffer containing all the data received during the
            simulation, and a flag indicating if any data was missing.

            .. note::
                Implementations should not assume that the total buffer is
                necessarily shorter than 1GB.

        :rtype: tuple(memoryview, bool)
        """
        try:
            with self.transaction() as cursor:
                region_id = self._get_region_id(cursor, x, y, p, region)
                data = self._read_contents(cursor, region_id)
                # TODO missing data
                return data, False
        except LookupError:
            return memoryview(b''), True

    def write_session_credentials_to_db(self):
        """
        Write Spalloc session credentials to the database if in use.
        """
        # pylint: disable=protected-access
        if not FecDataView.has_allocation_controller():
            return
        mac = FecDataView.get_allocation_controller()
        if mac.proxying:
            # This is now assumed to be a SpallocJobController;
            # can't check that because of import circularity.
            job = mac._job
            if isinstance(job, SpallocJob):
                with self.transaction() as cur:
                    job._write_session_credentials_to_db(cur)

    def _set_core_name(self, cursor, x, y, p, core_name):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :param str core_name:
        """
        try:
            cursor.execute(
                """
                INSERT INTO core (x, y, processor, core_name)
                VALUES (?, ?, ? ,?)
                """, (x, y, p, core_name))
        except sqlite3.IntegrityError:
            cursor.execute(
                """
                UPDATE core SET core_name = ?
                WHERE x = ? AND y = ? and processor = ?
                """, (core_name, x, y, p))

    def store_vertex_labels(self):
        with self.transaction() as cursor:
            for placement in FecDataView.iterate_placemements():
                self._set_core_name(cursor, placement.x, placement.y,
                                    placement.p, placement.vertex.label)
            for chip in FecDataView.get_machine().chips:
                for processor in chip.processors:
                    if processor.is_monitor:
                        self._set_core_name(
                            cursor, chip.x, chip.y, processor.processor_id,
                            f"SCAMP(OS)_{chip.x}:{chip.y}")

    def get_core_name(self, x, y, p):
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT core_name
                    FROM core
                    WHERE x = ? AND y = ? and processor = ?
                    """, (x, y, p)):
                return str(row["core_name"], 'utf8')



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bit_field_compressor_report import bitfield_compressor_report
from .bit_field_summary import BitFieldSummary
from .board_chip_report import board_chip_report
from .energy_report import EnergyReport
from .fixed_route_from_machine_report import fixed_route_from_machine_report
from .memory_map_on_host_chip_report import memory_map_on_host_chip_report
from .memory_map_on_host_report import memory_map_on_host_report
from .network_specification import network_specification
from .router_collision_potential_report import (
    router_collision_potential_report)
from .routing_table_from_machine_report import (
    routing_table_from_machine_report)
from .real_tags_report import tags_from_machine_report
from .write_json_machine import write_json_machine
from .write_json_placements import write_json_placements
from .write_json_routing_tables import write_json_routing_tables
from .drift_report import drift_report


__all__ = [
    "bitfield_compressor_report",
    "BitFieldSummary",
    "board_chip_report",
    "EnergyReport",
    "fixed_route_from_machine_report",
    "network_specification",
    "memory_map_on_host_chip_report",
    "memory_map_on_host_report",
    "router_collision_potential_report",
    "routing_table_from_machine_report",
    "tags_from_machine_report",
    "write_json_machine",
    "write_json_placements",
    "write_json_routing_tables",
    "drift_report"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import logging
import os
import re
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION)
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB

logger = FormatAdapter(logging.getLogger(__name__))

_DDL_FILE = os.path.join(os.path.dirname(__file__), "global.sql")
_RE = re.compile(r"(\d+)([_,:])(\d+)(?:\2(\d+))?")


class GlobalProvenance(SQLiteDB):
    """
    Specific implementation of the Database for SQLite 3.

    .. note::
        *Not thread safe on the same database file.*
        Threads can access different DBs just fine.

    .. note::
        This totally relies on the way SQLite's type affinities function.
        You can't port to a different database engine without a lot of work.
    """

    __slots__ = [
        "_database_file"
    ]

    @classmethod
    def get_global_provenace_path(cls):
        """
        Get the path of the current provenance database of the last run

        .. warning::
            Calling this method between start/reset and run may result in a
            path to a database not yet created.

        :raises ValueError:
            if the system is in a state where path can't be retrieved,
            for example before run is called
        """
        return os.path.join(
            FecDataView.get_timestamp_dir_path(),
            "global_provenance.sqlite3")

    def __init__(self, database_file=None, memory=False):
        """
        :param database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data.
            If omitted, either the default file path or an unshared in-memory
            database will be used (suitable only for testing).
        :type database_file: str or None
        :param bool memory:
            Flag to say unshared in-memory can be used.
            Otherwise a `None` file will mean the default should be used
        """
        if database_file is None and not memory:
            database_file = self.get_global_provenace_path()
        self._database_file = database_file
        SQLiteDB.__init__(self, database_file, ddl_file=_DDL_FILE,
                          row_factory=None, text_factory=None)

    def insert_version(self, description, the_value):
        """
        Inserts data into the version_provenance table

        :param str description: The package for which the version applies
        :param str the_value: The version to be recorded
        """
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO version_provenance(
                    description, the_value)
                VALUES(?, ?)
                """, [description, the_value])

    def insert_category(self, category, machine_on):
        """
        Inserts category into the category_timer_provenance  returning id

        :param TimerCategory category: Name of Category starting
        :param bool machine_on: If the machine was done during all
            or some of the time
        """
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO category_timer_provenance(
                    category, machine_on, n_run, n_loop)
                VALUES(?, ?, ?, ?)
                """,
                [category.category_name, machine_on,
                 FecDataView.get_run_number(),
                 FecDataView.get_run_step()])
            return cur.lastrowid

    def insert_category_timing(self, category_id, timedelta):
        """
        Inserts run time into the category

        :param int category_id: id of the Category finished
        :param ~datetime.timedelta timedelta: Time to be recorded
       """
        time_taken = (
                (timedelta.seconds * MICRO_TO_MILLISECOND_CONVERSION) +
                (timedelta.microseconds / MICRO_TO_MILLISECOND_CONVERSION))

        with self.transaction() as cur:
            cur.execute(
                """
                UPDATE category_timer_provenance
                SET
                    time_taken = ?
                WHERE category_id = ?
                """, (time_taken, category_id))

    def insert_timing(
            self, category, algorithm, work, timedelta, skip_reason):
        """
        Inserts algorithms run times into the timer_provenance table

        :param int category: Category Id of the Algorithm
        :param str algorithm: Algorithm name
        :param TimerWork work: Type of work being done
        :param ~datetime.timedelta timedelta: Time to be recorded
        :param skip_reason: The reason the algorithm was skipped or `None` if
            it was not skipped
        :type skip_reason: str or None
        """
        time_taken = (
                (timedelta.seconds * MICRO_TO_MILLISECOND_CONVERSION) +
                (timedelta.microseconds / MICRO_TO_MILLISECOND_CONVERSION))
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO timer_provenance(
                    category_id, algorithm, work, time_taken, skip_reason)
                VALUES(?, ?, ?, ?, ?)
                """,
                [category, algorithm, work.work_name, time_taken, skip_reason])

    def store_log(self, level, message, timestamp=None):
        """
        Stores log messages into the database

        :param int level:
        :param str message:
        """
        if timestamp is None:
            timestamp = datetime.now()
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO p_log_provenance(
                    timestamp, level, message)
                VALUES(?, ?, ?)
                """,
                [timestamp, level, message])

    def _test_log_locked(self, text):
        """
        THIS IS A TESTING METHOD.

        This will lock the database and then try to do a log
        """
        with self.transaction() as cur:
            # lock the database
            cur.execute(
                """
                INSERT INTO version_provenance(
                    description, the_value)
                VALUES("foo", "bar")
                """)
            # try logging and storing while locked.
            logger.warning(text)

    def run_query(self, query, params=()):
        """
        Opens a connection to the database, runs a query, extracts the results
        and closes the connection

        The return type depends on the use_sqlite_rows parameter.
        By default this method returns tuples (lookup by index) but the
        advanced tuple type can be used instead, which supports lookup by name
        used in the query (use ``AS name`` in the query to set).

        This method will not allow queries that change the database unless the
        read_only flag is set to False.

        .. note::
            This method is mainly provided as a support method for the later
            methods that return specific data. For new IntergationTests
            please add a specific method rather than call this directly.

        :param str query: The SQL query to be run. May include ``?`` wildcards
        :param ~collections.abc.Iterable(str or int) params:
            The values to replace the ``?`` wildcards with.
            The number and types must match what the query expects
        :param bool read_only: see :py:meth:`get_database_handle`
        :param bool use_sqlite_rows: see :py:meth:`get_database_handle`
        :return: A list possibly empty of tuples/rows
            (one for each row in the database)
            where the number and type of the values corresponds to the where
            statement
        :rtype: list(tuple or ~sqlite3.Row)
        """
        results = []
        with self.transaction() as cur:
            for row in cur.execute(query, params):
                results.append(row)
        return results

    def get_timer_provenance(self, algorithm):
        """
        Gets the timer provenance item(s) from the last run

        :param str algorithm:
            The value to LIKE search for in the algorithm column.
            Can be the full name, or have ``%``  and ``_`` wildcards.
        :return:
            A possibly multiline string with for each row which matches the
            like a line ``algorithm: value``
        :rtype: str
        """
        query = """
            SELECT algorithm, time_taken
            FROM timer_provenance
            WHERE algorithm LIKE ?
            """
        return "\n".join(
            f"{row[0]}: {row[1]}"
            for row in self.run_query(query, [algorithm]))

    def get_run_times(self):
        """
        Gets the algorithm running times from the last run. If an algorithm is
        invoked multiple times in the run, its times are summed.

        :return:
            A possibly multiline string with for each row which matches the
            like a line ``description_name: time``. The times are in seconds.
        :rtype: str
        """
        # We know the database actually stores microseconds for durations
        query = """
            SELECT description, SUM(time_taken) / 1000000.0
            FROM timer_provenance
            GROUP BY description
            ORDER BY the_value
            """
        return "\n".join(
            f"{row[0].replace('_', ' ')}: {row[1]} s"
            for row in self.run_query(query))

    def get_run_time_of_BufferExtractor(self):
        """
        Gets the buffer extractor provenance item(s) from the last run

        :return:
            A possibly multiline string with for each row which matches the
            ``LIKE %BufferExtractor``
        :rtype: str
        """
        return self.get_timer_provenance("%BufferExtractor")

    def get_category_timer_sum(self, category):
        """
        Get the total runtime for one category of algorithms

        :param TimerCategory category:
        :return: total off all run times with this category
        :rtype: int
        """
        query = """
             SELECT sum(time_taken)
             FROM category_timer_provenance
             WHERE category = ?
             """
        data = self.run_query(query, [category.category_name])
        try:
            info = data[0][0]
            if info is None:
                return 0
            return info
        except IndexError:
            return 0

    def get_category_timer_sums(self, category):
        """
        Get the runtime for one category of algorithms
        split machine on, machine off

        :param TimerCategory category:
        :return: total on and off time of instances with this category
        :rtype: int
        """
        on = 0
        off = 0
        query = """
             SELECT sum(time_taken), machine_on
             FROM category_timer_provenance
             WHERE category = ?
             GROUP BY machine_on
             """
        try:
            for data in self.run_query(query, [category.category_name]):
                if data[1]:
                    on = data[0]
                else:
                    off = data[0]
        except IndexError:
            pass
        return on, off

    def get_timer_sum_by_category(self, category):
        """
        Get the total runtime for one category of algorithms

        :param TimerCategory category:
        :return: total off all run times with this category
        :rtype: int
        """
        query = """
             SELECT sum(time_taken)
             FROM full_timer_view
             WHERE category = ?
             """
        data = self.run_query(query, [category.category_name])
        try:
            info = data[0][0]
            if info is None:
                return 0
            return info
        except IndexError:
            return 0

    def get_timer_sum_by_work(self, work):
        """
        Get the total runtime for one work type of algorithms

        :param TimerWork work:
        :return: total off all runtimes with this category
        :rtype: int
        """
        query = """
             SELECT sum(time_taken)
             FROM full_timer_view
             WHERE work = ?
             """
        data = self.run_query(query, [work.work_name])
        try:
            info = data[0][0]
            if info is None:
                return 0
            return info
        except IndexError:
            return 0

    def get_timer_sum_by_algorithm(self, algorithm):
        """
        Get the total runtime for one algorithm

        :param str algorithm:
        :return: total off all runtimes with this algorithm
        :rtype: int
        """
        query = """
             SELECT sum(time_taken)
             FROM timer_provenance
             WHERE algorithm = ?
             """
        data = self.run_query(query, [algorithm])
        try:
            info = data[0][0]
            if info is None:
                return 0
            return info
        except IndexError:
            return 0

    def retreive_log_messages(self, min_level=0):
        """
        Retrieves all log messages at or above the min_level

        :param int min_level:
        :rtype: list(tuple(int, str))
        """
        query = """
            SELECT message
            FROM p_log_provenance
            WHERE level >= ?
            """
        messages = self.run_query(query, [min_level])
        return list(map(lambda x: x[0], messages))



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .chip_power_monitor_machine_vertex import ChipPowerMonitorMachineVertex
from .command_sender import CommandSender
from .command_sender_machine_vertex import CommandSenderMachineVertex
from .data_speed_up_packet_gatherer_machine_vertex import (
    DataSpeedUpPacketGatherMachineVertex)
from .extra_monitor_support_machine_vertex import (
    ExtraMonitorSupportMachineVertex)
from .live_packet_gather import LivePacketGather
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex
from .multi_cast_command import MultiCastCommand
from .reverse_ip_tag_multi_cast_source import ReverseIpTagMultiCastSource
from .reverse_ip_tag_multicast_source_machine_vertex import (
    ReverseIPTagMulticastSourceMachineVertex)
from .streaming_context_manager import StreamingContextManager

__all__ = ["CommandSender", "CommandSenderMachineVertex",
           "ChipPowerMonitorMachineVertex",
           "DataSpeedUpPacketGatherMachineVertex",
           "ExtraMonitorSupportMachineVertex",
           "LivePacketGather", "LivePacketGatherMachineVertex",
           "MultiCastCommand", "ReverseIpTagMultiCastSource",
           "ReverseIPTagMulticastSourceMachineVertex",
           "StreamingContextManager"]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import logging
from time import sleep
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinnman.messages.sdp import SDPFlag, SDPHeader, SDPMessage
from spinnman.model.enums import CPUState
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SDP_RUNNING_MESSAGE_CODES)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))
_ONE_WORD = struct.Struct("<I")

_LIMIT = 10


def chip_provenance_updater(all_core_subsets):
    updater = _ChipProvenanceUpdater(all_core_subsets)
    # pylint: disable=protected-access
    updater._run()


class _ChipProvenanceUpdater(object):
    """
    Forces all cores to generate provenance data, and then exit.
    """

    __slots__ = ["__all_cores", "__app_id", "__txrx"]

    def __init__(self, all_core_subsets):
        """
        :param ~spinn_machine.CoreSubsets all_core_subsets:
        """
        self.__all_cores = all_core_subsets
        self.__app_id = FecDataView.get_app_id()
        self.__txrx = FecDataView.get_transceiver()

    def _run(self):
        # check that the right number of processors are in sync
        processors_completed = self.__txrx.get_core_state_count(
            self.__app_id, CPUState.FINISHED)
        total_processors = len(self.__all_cores)
        left_to_do_cores = total_processors - processors_completed

        progress = ProgressBar(
            left_to_do_cores,
            "Forcing error cores to generate provenance data")

        error_cores = self.__txrx.get_cores_in_state(
            self.__all_cores, CPUState.RUN_TIME_EXCEPTION)
        watchdog_cores = self.__txrx.get_cores_in_state(
            self.__all_cores, CPUState.WATCHDOG)
        idle_cores = self.__txrx.get_cores_in_state(
            self.__all_cores, CPUState.IDLE)

        if error_cores or watchdog_cores or idle_cores:
            raise ConfigurationException(
                "Some cores have crashed. "
                f"RTE cores {error_cores.values()}, "
                f"watch-dogged cores {watchdog_cores.values()}, "
                f"idle cores {idle_cores.values()}")

        # check that all cores are in the state FINISHED which shows that
        # the core has received the message and done provenance updating
        self._update_provenance(
            total_processors, processors_completed, progress)
        progress.end()

    def _update_provenance(
            self, total_processors, processors_completed, progress):
        """
        :param int total_processors:
        :param int processors_completed:
        :param ~.ProgressBar progress:
        """
        # pylint: disable=too-many-arguments
        left_to_do_cores = total_processors - processors_completed
        attempts = 0
        while processors_completed != total_processors and attempts < _LIMIT:
            attempts += 1
            unsuccessful_cores = self.__txrx.get_cores_not_in_state(
                self.__all_cores, CPUState.FINISHED)

            for (x, y, p) in unsuccessful_cores.keys():
                self._send_chip_update_provenance_and_exit(x, y, p)

            processors_completed = self.__txrx.get_core_state_count(
                self.__app_id, CPUState.FINISHED)

            left_over_now = total_processors - processors_completed
            to_update = left_to_do_cores - left_over_now
            left_to_do_cores = left_over_now
            if to_update != 0:
                progress.update(to_update)
                sleep(0.5)
        if processors_completed != total_processors:
            logger.error("Unable to Finish getting provenance data. "
                         "Abandoned after too many retries. "
                         "Board may be left in an unstable state!")

    def _send_chip_update_provenance_and_exit(self, x, y, p):
        """
        :param int x:
        :param int y:
        :param int p:
        """
        cmd = SDP_RUNNING_MESSAGE_CODES.SDP_UPDATE_PROVENCE_REGION_AND_EXIT
        port = SDP_PORTS.RUNNING_COMMAND_SDP_PORT

        self.__txrx.send_sdp_message(SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED,
                destination_port=port.value, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y),
            data=_ONE_WORD.pack(cmd.value)))



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsBitFieldRoutingCompression(
        object, metaclass=AbstractBase):
    """
    Marks a machine vertex that can support having the on-chip bitfield
    compressor running on its core.
    """
    __slots__ = ()

    @abstractmethod
    def bit_field_base_address(self, placement):
        """
        Returns the SDRAM address for the bit-field table data.

        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield address
        :rtype: int
        """

    @abstractmethod
    def regeneratable_sdram_blocks_and_sizes(self, placement):
        """
        Returns the SDRAM addresses and sizes for the cores' SDRAM that
        are available (borrowed) for generating bitfield tables.

        :param ~pacman.model.placements.Placement placement:
        :return: list of tuples containing (the SDRAM address for the cores
            SDRAM address's for the core's SDRAM that can be used to generate
            bitfield tables loaded, and the size of memory chunks located
            there)
        :rtype: list(tuple(int,int))
        """
